from pydantic_settings import BaseSettings, SettingsConfigDict


class DbSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="NC_", env_file=".env", extra="ignore")
    db_host: str
    db_port: int
    db_user: str
    db_password: str
    db_name: str
    db_schema: str
    db_pool_min: int = 1
    db_pool_max: int = 5

    @classmethod
    def from_env(cls):
        """
        This is a convenience method that prevents lint errors when variables are set from the environment.
        """
        return cls()  # type: ignore




from typing import Any, Dict, Type

# NOTE: This module should not have any dependencies on any other nc-db modules

VERSION = 1
VERSION_TABLE = 'db_version'
CLASS_REGISTRATION_TABLE = 'registered_classes'
CLASS_VERSION_DICT: Dict[Type[Any], int] = {}
ALLOW_SCHEMA_DRIFT: set[Type[Any]] = set()



import asyncio
from string.templatelib import Template
from typing import Any, List
from psycopg_pool import AsyncConnectionPool
from psycopg import AsyncConnection, sql
from nc_db.metadata import nc_config
from nc_db.metadata.class_metadata import ClassMetadata
from nc_db.handlers.common import table_exists, get_config_version

from nc_db.identifiable import Identifiable

# The number doesn't really matter -- we just need to ensure every instance of the
# class uses the same lock number.
DB_CONFIG_LOCK = 100
"""
A pg advisory lock used to guarantee only one process creates the config tables.
"""


class CommandHandler():
    """
    Handles executing commands against the database
    """

    def __init__(self, connection_pool: AsyncConnectionPool):
        self._pool = connection_pool

    @classmethod
    async def set_search_path(cls, schema: str, conn: AsyncConnection[Any]):
        # autopep8: off
        await conn.execute(t"SET search_path TO {schema:i}, public")
        # autopep8: on
        await conn.commit()

    async def ensure_config_tables(self):
        """
        Ensures the configuration tables exist in the database.
        If the tables exist it pulls the version number from the database.
        If the tables do not exist it creates them and adds the version into the table.

        Returns: True if the version in the config table matches the nc-db library version in nc_config.py

        :param self: Description
        """

        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:
                # autopep8: off
                await cursor.execute(t"SELECT pg_advisory_xact_lock({DB_CONFIG_LOCK})")
                # autopep8: on
                if not await table_exists(cursor, nc_config.VERSION_TABLE):

                    # autopep8: off
                    # 1. Create the version table
                    create_version_table = t"""
                        CREATE TABLE {nc_config.VERSION_TABLE:i} (
                            version INT PRIMARY KEY
                        );
                        """

                    # 2. Insert the version
                    insert_version = t"INSERT INTO {nc_config.VERSION_TABLE:i} VALUES ({nc_config.VERSION});"

                    # 3. Create the schema_config table to make sure class schemas don't change without the
                    #       schema version number being updated.
                    create_registration_table = t"""
                        CREATE TABLE {nc_config.CLASS_REGISTRATION_TABLE:i} (
                            name TEXT PRIMARY KEY,
                            signature TEXT
                        );
                        """
                    # autopep8: on
                    await cursor.execute(create_version_table)
                    await cursor.execute(insert_version)
                    await cursor.execute(create_registration_table)
                db_version = await get_config_version(cursor, nc_config.VERSION_TABLE)
                if (db_version != nc_config.VERSION):
                    raise RuntimeError("It looks like this database was used with an older version of nc-db (version mismatch)")

    async def ensure_class_table(self, metadata: ClassMetadata) -> None:
        """
        A thread-safe function that creates a table for storing objects of the class defined by metadata if
        the table doesn't already exist.

        :param metadata: Metdata representing a class whose instances will be stored in the database.
        """
        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:
                # Get the table name for this class
                table_name = metadata.table_name
                # autopep8: off
                # To ensure there are no issue with concurrency, we get a transaction-based advisory lock so 
                # only one process / task can be in this section at a time.
                advisory_lock = t"SELECT pg_advisory_xact_lock(hashtext({metadata.table_name}))"
                await cursor.execute(advisory_lock)
                if not await table_exists(cursor,metadata.table_name):
                    # 1. Create a table to store instances of this class
                    columns = [t"{col.name:i} {Template(col.column_definition):q}" 
                                for col in 
                                metadata.columns
                                ]
                    column_definitions = sql.SQL(",").join(columns)
                    create_table=t"""
                        CREATE TABLE {table_name:i} ({column_definitions:q});
                        """
                    await cursor.execute(create_table)
                    create_indexes = [
                        t"CREATE INDEX {f'{col.name}_index':i} on {metadata.table_name:i} ({col.name:i});"
                        for col in metadata.columns
                        if col.index and not col.primary_key
                    ]
                    index_tasks = [ cursor.execute(index) for index in create_indexes]
                    await asyncio.gather(*index_tasks)

                    # 2. Create an entry in the class registration table with this class signature.
                    insert_signature = t"INSERT INTO {nc_config.CLASS_REGISTRATION_TABLE:i} VALUES ({metadata.table_name}, {metadata.signature});"
                    await cursor.execute(insert_signature)
                    # autopep8: on

    async def insert(self, metadata: ClassMetadata, obj: Identifiable) -> int:
        """
        Saves the object to the database, updates the id attribute with the new id, and returns the id.
        """
        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:
                # autopep8: off
                # Create SQL for all columns
                obj_dict = obj.model_dump()
                # The following call updates the obj_dict (e.g. removes data stored in columns) and gets a list of db updates
                db_updates = [col.object_to_db_update(obj_dict) for col in metadata.columns]
                columns = [t"{db_update.column_name:i}" for db_update in db_updates if db_update]
                values = [t"{db_update.value}" for db_update in db_updates if db_update is not None]
                columns_sql = sql.SQL(",").join(columns)
                values_sql = sql.SQL(",").join(values)
                insert_cmd = t"""INSERT INTO {metadata.table_name:i} ({columns_sql:q}) VALUES (
                    {values_sql:q}
                )  RETURNING id;"""

                # autopep8: on
                await cursor.execute(insert_cmd)
                row = await cursor.fetchone()
                if row:
                    id: int = row[0]
                    obj.id = id
                    return id
                else:
                    raise RuntimeError(
                        f"Could not insert object of type {metadata.class_name}. The database did not return a new id.")

    async def update(self, metadata: ClassMetadata, obj: Identifiable):
        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:

                # autopep8: off
                # Create SQL to set values for all columns
                obj_dict = obj.model_dump()
                db_updates = [col.object_to_db_update(obj_dict) for col in metadata.columns]
                set_cmds = [t"{db_update.column_name:i}={db_update.value}" for db_update in db_updates if db_update ]
                set_sql = sql.SQL(",").join(set_cmds)
                update_cmd = t"UPDATE {metadata.table_name:i} SET {set_sql:q} WHERE id={obj.id};"

                # autopep8: on
                await cursor.execute(update_cmd)

    async def delete(self, metadata: ClassMetadata, ids: List[int]) -> List[int]:
        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:
                # autopep8: off
                delete_cmd=t"DELETE FROM {metadata.table_name:i} WHERE id=ANY({ids}) RETURNING id;"
                # autopep8: on
                await cursor.execute(delete_cmd)
                results = await cursor.fetchall()
                return [result[0] for result in results]



from typing import Annotated, TypeVar
from pydantic import BaseModel, Field

from nc_db.metadata.db_column import DbColumn


NEW_ID = -1

"""
A new entity should always be initialized with this value
"""


class Identifiable(BaseModel):
    """
    Represents the base properties needed to uniquely identify objects stored in the database
    """
    id: Annotated[int, DbColumn('INT GENERATED ALWAYS AS IDENTITY PRIMARY KEY', True, True, True)] = Field(default=NEW_ID)


T = TypeVar('T', bound=Identifiable)



"""
This folder contains very small DB functions that are reused across queries and commands
"""
from psycopg import AsyncCursor


async def table_exists(cursor: AsyncCursor, table: str):
    TABLES_EXISTS_INDEX = 0
     # autopep8: off -- it doesn't support t-strings yet
    query = t"""
        SELECT to_regclass({table:t}) IS NOT NULL AS table_exists;

    """
    # autopep8: on
    await cursor.execute(query)
    row = await cursor.fetchone()
    if row:
        return row[TABLES_EXISTS_INDEX] == 1
    else:
        return False


async def get_config_version(cursor: AsyncCursor, config_table_name: str) -> int:
     # autopep8: off -- it doesn't support t-strings yet
    query = t"SELECT version from {config_table_name:i}"
    # autopep8: on
    row = await cursor.execute(query)
    record = await row.fetchone()
    if record:
        return int(record[0])
    else:
        raise RuntimeError("No nc_db version number found in the database.")


async def get_registered_signature(cursor: AsyncCursor, class_name: str, registration_table_name: str) -> str:
     # autopep8: off -- it doesn't support t-strings yet
    query = t"SELECT signature from {registration_table_name:i} where name={class_name}"
    # autopep8: on
    row = await cursor.execute(query)
    record = await row.fetchone()
    if record:
        return str(record[0])
    else:
        raise RuntimeError("No class registration found in the database.")





import re


POSTGRES_RESERVED_WORDS = {
    "ALL",
    "ANALYSE",
    "ANALYZE",
    "AND",
    "ANY",
    "ARRAY",
    "AS",
    "ASC",
    "ASYMMETRIC",
    "AUTHORIZATION",
    "BINARY",
    "BOTH",
    "CASE",
    "CAST",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CURRENT_CATALOG",
    "CURRENT_DATE",
    "CURRENT_ROLE",
    "CURRENT_SCHEMA",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "DEFAULT",
    "DEFERRABLE",
    "DESC",
    "DISTINCT",
    "DO",
    "ELSE",
    "END",
    "EXCEPT",
    "FALSE",
    "FETCH",
    "FOR",
    "FOREIGN",
    "FREEZE",
    "FROM",
    "FULL",
    "GRANT",
    "GROUP",
    "HAVING",
    "ILIKE",
    "IN",
    "INITIALLY",
    "INNER",
    "INTERSECT",
    "INTO",
    "IS",
    "ISNULL",
    "JOIN",
    "LATERAL",
    "LEADING",
    "LEFT",
    "LIKE",
    "LIMIT",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "NATURAL",
    "NOT",
    "NOTNULL",
    "NULL",
    "OFFSET",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "OUTER",
    "OVERLAPS",
    "PLACING",
    "PRIMARY",
    "REFERENCES",
    "RETURNING",
    "RIGHT",
    "SELECT",
    "SESSION_USER",
    "SIMILAR",
    "SOME",
    "SYMMETRIC",
    "TABLE",
    "THEN",
    "TO",
    "TRAILING",
    "TRUE",
    "UNION",
    "UNIQUE",
    "USER",
    "USING",
    "VARIADIC",
    "VERBOSE",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
}


def to_pg_identifier(identifier: str) -> str:
    """
    Makes sure a string isn't a pg reserved word and converts it to snake_case.
    Throws an exception if the identifier is a Postgres reserved word
    :identifier: A string in pascal or camel case
    :rtype: str
    """
    """
    Convert PascalCase or camelCase to lower_snake_case.

    Examples:
        "PascalCase"      -> "pascal_case"
        "camelCase"       -> "camel_case"
        "HTTPRequest"     -> "http_request"
        "JSONDataParser"  -> "json_data_parser"
        "myURLValue"      -> "my_url_value"
    """
    POSTGRES_IDENTIFIER_LENGTH_LIMIT = 63
    # Handle acronym-to-word boundaries (JSONData -> JSON_Data)
    s1 = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", identifier)

    # Handle lower/number to upper (camelCase, myURLValue)
    s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1)

    valid_identifier = s2.lower()
    if valid_identifier.upper() in POSTGRES_RESERVED_WORDS:
        raise RuntimeError("Class name must not be a Postgres reserved word")
    if len(valid_identifier) > POSTGRES_IDENTIFIER_LENGTH_LIMIT:
        raise RuntimeError(f"Class name exceeds {POSTGRES_IDENTIFIER_LENGTH_LIMIT} characters. Class {identifier} is invalid.")
    return valid_identifier




from typing import Any, Dict
from psycopg.types.json import Jsonb

from nc_db.metadata.column_mappers.column_mapper import ColumnMapper, DbUpdate


class BodyMapper(ColumnMapper):
    """
    The body column requires special logic as it serializes and deserializes all class atributes that are not stored as a column.
    This class encapsulates that logic
    NOTE: This mapper MUST RUN LAST as the other mappers may modify the object dictionary (e.g. removing keys for values stored in
        other columns).
    """

    def __init__(self):
        super().__init__('body', 'JSONB', False, False)

    def db_row_to_object[T](self, from_db_row: Dict[str, Any], to_object_dict: Dict[str, Any]):
        #  Get a dictionary of all class attributes that aren't stored in their own column
        # and update the to_object_dict with it
        # NOTE: Custom select statements may query for results that don't have a body
        # column. Everything else should have one.
        partial_object_dict: Dict[str, Any] | None = from_db_row.get(self._name)
        if partial_object_dict is not None:
            to_object_dict.update(partial_object_dict)

    def object_to_db_update(self, obj_dict: Dict[str, Any]) -> DbUpdate:
        # Get the obj_dict and serialize it to jsonb
        # NOTE: This mapper must run last. It assumes all attributes that should be stored
        #   in their own columns have already been removed from obj_dict
        return DbUpdate(
            value=Jsonb(obj_dict),
            column_name=self.name
        )




from dataclasses import dataclass
from typing import Any, Dict


@dataclass(frozen=True)
class DbUpdate():
    column_name: str
    """
    The name of the column this update applies to.
    """

    value: Any = None
    """
    The value the column should be updated to in the DB column.
    """


class ColumnMapper():
    """
    The base column mapper contains data and functions for mapping class attributes into columns in
    the database. Thi class can be derived from for special columns (e.g. body and signature).
    NOTE: Most attributes are stored as jsonb in the body column and don't need a column mapper. Only
    class attributes annotated with DbColumn will be mapped to their own columns.
    """
    @property
    def name(self) -> str:
        """
        The name of the column. This MUST match the attribute name and must be lowercase (e.g. snake_case)
        """
        return self._name

    @property
    def column_definition(self) -> str:
        """
        A SQL string that defines the column.
        """
        return self._column_definition

    @property
    def index(self) -> bool:
        """
        Indicates if the column is an index
        """
        return self._index

    @property
    def auto_generated(self) -> bool:
        """
        Indicates if the DB generates the value. If True, the column is excluded on inserts/updates.
        """
        return self._auto_generated

    @property
    def primary_key(self) -> bool:
        """
        This column is the primary key if True.
        """
        return self._primary_key

    def __init__(self, name: str, column_definition: str, index: bool = False, auto_generated: bool = False, primary_key: bool = False):
        self._name = name
        self._column_definition = column_definition
        self._index = index
        self._auto_generated = auto_generated
        self._primary_key = primary_key

    def db_row_to_object[T](self, from_db_row: Dict[str, Any], to_object_dict: Dict[str, Any]):
        """
        Takes the row received from the db and adds a key and value to the object_dict that correspond
        to this column's value from the row.

        :param from_db_row: A row retrieved from the DB containing a value for this column.
        :param to_object_dict: A dictionary that will be used to construct an object using pydantic's model_validate function
        """
        to_object_dict[self._name] = from_db_row[self.name]

    def object_to_db_update(self, obj_dict: Dict[str, Any]) -> DbUpdate | None:
        """
        Takes a dictionary representing an object, removes the key from it (if appropriate) and returns
        a SetCommand for the object. the value for this column. This function should
        modify the obj_dict if required. For most columns, this will delete the column's key from the obj_dict

        :param obj_dict: A dictionary representing the object's attributes. This comes from pydantic's model_dump function and
            will be used to serialize the object into the body column (jsonb). Most column values should be removed from this prior
            to serializing to jsonb.
        :rtype: The value from this object's dictionary that should be stored in this column in the DB
        """
        column_value = obj_dict[self.name]
        del obj_dict[self.name]
        if self.auto_generated:
            return None
        else:
            return DbUpdate(
                self.name,
                value=column_value,
            )



from dataclasses import dataclass
from typing import Annotated, TypeAlias

# NOTE: This class should not have any dependencies on other nc-db modules.


@dataclass(frozen=True)
class DbColumn():
    """
    Description
    -----------
    Required for any attributes you wish to index/search in the database. Specifies indexes and column type.
    You can use pre-canned type aliases for convenience (IntDbIndex, StrDbIndex, etc.) or define your own.

    column_definition: Must be a valid postgres column definition ('INT', 'VARCHAR UNIQUE','INT GENERATED ALWAYS', etc.)

    Examples:
    ---------
    - Using Annotated for full customization
    class Foo():
        my_int: Annotated[int,DbIndex('BIGINT UNIQUE NOT NULL']

    - Using pre-canned type aliases for simple use cases
    class Foo():
        my_int: IntDbIndex
    """

    column_definition: str
    """
    The postgres column definition.
    Examples:
        'INT UNIQUE'
        'VARCHAR NOT NOLL'
        'INT GENERATED ALWAYS AS IDENTITY PRIMARY KEY'
    """

    index: bool = False
    """
    Indicates whether an index should be created for this column.
    """

    auto_generated: bool = False
    """
    If set to TRUE, this attribute is ignored when saving to the DB. When
    retrieving, it's value will be set from the corresponding column in the DB.
    You MUST set GENERATED ALWAYS in your column descriprion if this is true.
    """

    primary_key: bool = False
    """
    If set to True, the nc-db library does not create an index for this column.
    This attribute is the primary key for the table--you MUST specify
    "PRIMARY KEY along with the type, etc in the column_definition."
    """


IntDbIndex: TypeAlias = Annotated[int, DbColumn('INT', True, False)]
BigIntDbIndex: TypeAlias = Annotated[int, DbColumn('BIGINT', True, False)]
StrDbIndex: TypeAlias = Annotated[str, DbColumn('TEXT', True, False)]
FloatDbIndex: TypeAlias = Annotated[float, DbColumn('FLOAT', True, False)]



from typing import Callable, Type
from nc_db.metadata.nc_config import CLASS_VERSION_DICT, ALLOW_SCHEMA_DRIFT


def schema_version[T](version: int, allow_schema_drift: bool = False)->Callable[[Type[T]], Type[T]]:
    """
    Specifies the schema version of the class.
    MUST be monotonically increasing.
    MUST be updated when any attributes with the DbColumn annotation are created/modified/deleted.
    May by updated for any other reason, causing class instances to be stored/retrieved from a new table.

    Parameters:
    version: The version of the schema
    allow_schema_drift: If True, errors will only be thrown when class/schema changes break functionality.
                        Default (recommended) is False which will throw errors when any changes are made to the class and the
                        schema_version isn't updated.
    """
    def decorator(cls: Type[T]):
        CLASS_VERSION_DICT.update({cls: version})
        if allow_schema_drift:
            ALLOW_SCHEMA_DRIFT.update({cls: True})
        return cls
    return decorator




import hashlib
import json
from typing import Annotated, Any, Dict, List, Type, cast, get_args, get_origin, get_type_hints
from nc_db.identifiable import Identifiable
from nc_db.metadata.column_mappers.body_mapper import BodyMapper
from nc_db.metadata.column_mappers.column_mapper import ColumnMapper
from nc_db.metadata.db_column import DbColumn
import base64

from nc_db.metadata.pg_utils import to_pg_identifier


class ClassMetadata():
    """
    Contains metadata statically derived through class introspection.
    IMPORTANT: Every property in this class must be deterministically set based upon the cls property passed
    to the constructor. The constructor MUST be idempotent.

    This class extracts metadata from a class type. It does not require any connectivity or
    data from the database.
    """
    @property
    def cls(self) -> Type[Identifiable]:
        """
        The Python class definition
        """
        return self._cls

    @property
    def table_name(self) -> str:
        """
        The table name where instances of this class will be stored in the DB.
        """
        return self._table_name

    @property
    def schema_version(self) -> int:
        """
        The schema version of the class. This is user-defined and should be monotonic.
        MUST be incremented when a class is updated with new / modified / deleted DbColumn attributes.
        """
        return self._schema_version

    @property
    def columns(self) -> List[ColumnMapper]:
        """
        Used to generate b-tree indexes in Postgres
        """
        return self._columns

    @property
    def signature(self) -> str:
        """
        A unique signature for this class based ONLY upon its schema and indexes.
        NOTE: Over time, a class may be modified. Each time this occurs, the unique schema identifier will
        change.
        NOTE: This signature must be deterministic based only upon the structure of the class . It must not
        depend upon any data from the DB or other sources.
        """
        return self._signature

    @property
    def class_name(self):
        """
        The friendly, easy-to-read name of this class.
        :param self: Description
        """
        return self._name

    @property
    def mutable_columns(self) -> List[ColumnMapper]:
        """
        Returns all columns that are not GENERATED ALWAYS by the db
        """
        return [column for column in self.columns if not column.auto_generated]

    @property
    def immutable_columns(self) -> List[ColumnMapper]:
        """
        Returns all columns that are GENERATED ALWAYS by the db
        """
        return [column for column in self.columns if column.auto_generated]

    def __init__(
            self, cls: Type[Identifiable],
            schema_version: int
    ):
        self._cls = cls
        self._name = cls.__name__
        self._schema_version = schema_version
        self._table_name = to_pg_identifier(f"{self._name}_{self._schema_version}")
        self._columns: List[ColumnMapper] = []
        full_hints = get_type_hints(cls, include_extras=True)
        # Add user-defined columns
        for attr_name, type_hint in full_hints.items():
            origin = get_origin(type_hint)
            if origin is Annotated:

                # The first arg is the type -- we already know what this is, so exclude it.
                args = get_args(type_hint)
                indexes = [arg for arg in args if isinstance(arg, DbColumn)]
                index_count = len(indexes)
                if index_count > 1:
                    raise RuntimeError(
                        "Multiple DbColumn instances where annotated on a single class attribute. Only one DbIndex instance is allowed per attribute.")
                if index_count == 1:
                    if attr_name != attr_name.lower():
                        raise RuntimeError("All class attributes that have an index or column specified must be lowercase (e.g. snake_case)")
                    db_index = indexes[0]
                    self.columns.append(ColumnMapper(attr_name, db_index.column_definition, True, db_index.auto_generated, db_index.primary_key))
        # Generate the signature now that we have the user-defined columns available.
        self._signature = self._generate_signature(cls)

        # All classes need a body acolumn. The body column MUST be last to allow the other column mappers to
        # remove their value from the dictionary. This is done to reduce storage and ensure data consistency (no need to store
        # values in the json and their own column)
        self.columns.append(BodyMapper())

    def _generate_signature(self, cls: Type[Identifiable]) -> str:
        """
        Deterministically generates a signature based solely upon name and structure of
        the class.
        """
        schema = cls.model_json_schema()

        def clean_schema(node: Any) -> None:
            """
            Recursively removes metadata that doesn't change the structure of a model.
            This is removed so we can autogenerate a version (hash) to ensure we
            don't mix versions of the same class name.
            Adds a string representing the indexes so new indexes can be added and migrations supported.

            :param node: A JSON node
            """
            keys_to_remove = ['title', 'description', 'example']
            if isinstance(node, dict):
                node = cast(Dict[str, Any], node)
                # Remove metadata keys that don't affect the data contract
                for key in keys_to_remove:
                    if key in node:
                        del node[key]
                for value in list(node.values()):
                    clean_schema(value)
            elif isinstance(node, list):
                node = cast(List[Any], node)
                for item in node:
                    clean_schema(item)
            return node
        clean_schema(schema)

        class_schema = json.dumps(schema, sort_keys=True, default=str, ensure_ascii=True)
        # Add all of the column names into the hash so we get a unique hash if any of the columns or indexes change.
        for column in self.columns:
            class_schema += f"___{column.name}"
        hash = hashlib.sha256(class_schema.encode('utf-8'), usedforsecurity=False)
        # To minimize characers, encode as base 32 and remove padding ('=')
        signature = base64.b32encode(hash.digest()).decode('utf-8').replace("=", "")
        return signature



from string.templatelib import Template
from typing import Any, Dict, List, Optional, Type
from psycopg_pool import AsyncConnectionPool
from psycopg.rows import dict_row
from nc_db.metadata import nc_config
from nc_db.handlers.common import get_registered_signature as _get_registered_signature, table_exists as _table_exists, get_config_version as _get_config_version
from nc_db.identifiable import Identifiable
from nc_db.metadata.class_metadata import ClassMetadata


class QueryHandler():
    """
    Handles executing queries against the database
    """

    def __init__(self, connection_pool: AsyncConnectionPool):
        self._pool = connection_pool

    def _create_objects[T:Identifiable](self, cls: Type[T], metadata: ClassMetadata, rows: List[Dict[str, Any]] | None) -> List[T]:
        objects: List[T] = []
        if rows:
            for row in rows:
                # Create a blank dictionary and have each column mapper populate it
                obj_dict: Dict[str, Any] = {}
                for column in metadata.columns:
                    column.db_row_to_object(row, obj_dict)
                # Use pydantic's model_validate to create a class instance from the dictionary.
                objects.append(cls.model_validate(obj_dict))
        return objects

    async def table_exists(self, table: str) -> bool:
        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:
                return await _table_exists(cursor, table)

    async def get_db_config_version(self) -> int:
        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:
                return await _get_config_version(cursor, nc_config.VERSION_TABLE)

    async def get_registered_signature(self, metadata: ClassMetadata) -> str:
        async with self._pool.connection() as conn:
            async with conn.cursor() as cursor:
                return await _get_registered_signature(cursor, metadata.table_name, nc_config.CLASS_REGISTRATION_TABLE)

    async def get[T: Identifiable](self, cls: Type[T], metadata: ClassMetadata, ids: Optional[List[int]] = None) -> List[T]:
        async with self._pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                # autopep8: off
                if ids:
                    query = t"SELECT * FROM {metadata.table_name:i} WHERE id=ANY({ids})"
                else:
                    query = t"SELECT * FROM {metadata.table_name:i}"
                # autopep8: on
                await cursor.execute(query)
                rows = await cursor.fetchall()
                result = self._create_objects(cls, metadata, rows)
                return result

    async def search[T:Identifiable](self, cls: Type[T], metadata: ClassMetadata, search_condition_template: Template) -> List[T]:
        async with self._pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                # autopep8: off
                query = t"SELECT * FROM {metadata.table_name:i} WHERE "
                # autopep8: on
                query += search_condition_template
                await cursor.execute(query)
                rows = await cursor.fetchall()
                result = self._create_objects(cls, metadata, rows)
                return result

    async def select[T:Identifiable](self, cls: Type[T], metadata: ClassMetadata, select_sql: Template) -> List[T]:
        async with self._pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(select_sql)
                rows = await cursor.fetchall()
                result = self._create_objects(cls, metadata, rows)
                return result



import asyncio
from typing import Any, Callable, Dict, List, Optional, Type, overload
from psycopg import AsyncConnection
from psycopg_pool import AsyncConnectionPool

from nc_db.db_settings import DbSettings
from nc_db.metadata.class_metadata import ClassMetadata
from nc_db.handlers.query_handler import QueryHandler
from nc_db.handlers.command_handler import CommandHandler
from nc_db.identifiable import NEW_ID, T, Identifiable
from string.templatelib import Template

from nc_db.metadata.nc_config import CLASS_VERSION_DICT, ALLOW_SCHEMA_DRIFT


class Db:
    """
    Provides a document-oriented DB system using Postgres with b-tree indexes for better performance.
    Think mongodb for Postgres with some help from Pydantic.
    New collections (tables) are created automatically but model and schema migrations are not (use the migrate function or roll your own).
    Collections are defined by a class's signature. All objects in a collection are guaranteed to have the same shape (serialize and deserialize).
    A class's signature is unique based upon the structure of the class and the indexes it defines.
    To improve performance with large objects:
        1. Each object type has a dedicated table in the database.
        1. Class attributes annotated with DbIndex are automatically maintained in dedicated columns (B-trees, no GIN indexes)
        1. Attributes you want to search on MUST be annotated with DbIndex
            1. You can use the DbIndex type aliases for common use cases
            Examples:
                - my_class_var: Annotated[int,DbIndex('INT',unique=True, nullable=False)]
                - my_class_var: DbIntIndex

    The Db instance should share a lifecycle with your service or application. Recommended practice is to treat it like
    a singleton for each database your service connects to. It is asynchronous and uses a connection pool under the hood.
    """

    def __init__(self, config: DbSettings | AsyncConnectionPool):
        """
        Creates a Db instance that can be used for CRUD against a Postgres database.

        :param config: Either the DbSettings or a connection pool. If DbSettings, the Db class will create the connection pool.
        """
        self._connected = False
        if isinstance(config, DbSettings):
            self._pool = Db.create_connection_pool(config)
        else:
            self._pool = config
        self._class_metadata_dict: Dict[Type[Identifiable], ClassMetadata] = {}

    async def connect(self):
        """
        Connects the Db class to the database and ensures the DB is configured correctly.
        """
        if not self._connected:
            await self._pool.open()
            self._query_handler = QueryHandler(self._pool)
            self._command_handler = CommandHandler(self._pool)

            # 1. Ensure the config tables exist
            await self._command_handler.ensure_config_tables()

            self._connected = True

    async def disconnect(self):
        """
        Closes all connections to the DB
        """
        await self._pool.close()
        self._connected = False

    @classmethod
    def create_connection_pool(cls, settings: DbSettings):
        """
        A utility function that creates a connection pool for the database specified
        by the settings parameter. This is called automatically by the Db.connect() function

        settings: Specifies the information needed to create the connection pool to a specific database
        """

        #
        async def configure_connection(conn: AsyncConnection[Any]):
            # This is the only place the search path should be set.
            # We set it here to configure every connection up-front.
            await CommandHandler.set_search_path(settings.db_schema, conn)

        # Build DSN string
        dsn = (
            f"postgresql://{settings.db_user}:{settings.db_password}"
            f"@{settings.db_host}:{settings.db_port}/{settings.db_name}"
        )
        pool: AsyncConnectionPool = AsyncConnectionPool(
            conninfo=dsn,
            min_size=settings.db_pool_min,
            max_size=settings.db_pool_max,
            open=False,
            configure=configure_connection
        )
        return pool

    def _ensure_connection(self, raise_if_false: bool = True) -> bool:
        if self._connected:
            return True
        elif raise_if_false:
            raise RuntimeError("Not connected to the DB. You must await Db.connect() prior to making calls that interact with the DB")
        return False

    async def _get_class_metadata(self, cls: Type[T]) -> ClassMetadata:
        """
        Gets / creates metadata for cls and ensures the appropriate database tables exist for this class.
        Returns the metadata object once the DB is configured to support CRUD on the class represented by cls.

        :param cls: The class to be verified
        Returns metadata for the class represented by cls.
        """
        # Make sure we're connected to the database
        self._ensure_connection()

        version = CLASS_VERSION_DICT.get(cls)
        if version is None:
            raise RuntimeError(
                f"Class {cls.__name__} does not have the @schema_version decorator. You must specify a schema version using the @schema_version class decorator")

        # If we have DbClassMetadata, return it, otherwise
        # get / create it
        metadata = self._class_metadata_dict.get(cls, None)
        if metadata is None:
            # NOTE: We don't need to have a critical section here as the ClassMetadata constructor is idempotent
            # and the command_handler.ensure_class_table function uses an advisory lock to create the table if
            # it doesn't exist.
            # The metadata_dict is effectively just a cache to avoid reprocessing the cls type multiple times.

            # NOTE: The (very minor) downside to not using a critical section here is that, if multiple calls are made simultaneously
            # the very first time we see a class, we will check the DB for each

            # We don't need to check for the key as the Identifiable class throws an error at import if derived classes
            # haven't specified a schema_version with the @schema_version decorator.
            version = CLASS_VERSION_DICT[cls]
            metadata = ClassMetadata(cls, version)
            await self._command_handler.ensure_class_table(metadata)
            if cls not in ALLOW_SCHEMA_DRIFT:
                registered_signature = await self._query_handler.get_registered_signature(metadata)
                if registered_signature != metadata.signature:
                    # autopep8: off
                    msg = f"""Run-time class {metadata.class_name} version {metadata.schema_version} has a signature ({metadata.signature}) that doesn't match the registered signature in the database ({registered_signature}) """
                    # autopep8: on
                    raise ValueError(msg)
            self._class_metadata_dict.update({cls: metadata})
        return metadata

    async def insert(self, item: Identifiable) -> int:
        """
        Inserts a new instance of an object into the database.
        The object must meet the following criteria:
            1. id must be equal to NEW_ID (call update if this is an update to an existing object)
            1. The ClassConfig for this object type (class) must have already been registered once (and only once).
        NOTE: The object's id will be set to match the new id provided by the database.


        :param object: A new object to be stored in the database
        :return: The new id for the object. NOTE: After calling this function, the obect's id will be set to match this value.
        """
        if item.id != NEW_ID:
            raise ValueError("When calling insert, the object's id must be NEW_ID. Call update to update an existing object.")
        metadata = await self._get_class_metadata(item.__class__)

        result = await self._command_handler.insert(metadata, item)
        return result

    @overload
    async def search(self, cls: Type[T], search_condition: str, value: Any,
                     ) -> List[T]: ...

    @overload
    async def search(self, cls: Type[T], search_condition: Template) -> List[T]: ...

    async def search(self,
                     cls: Type[T],
                     search_condition: str | Template,
                     value: Optional[Any] = None) -> List[T]:
        """
        Description
        -----------
        Searches for objects in the database. All attributes used in the search must be declared of type DbIndex[T]

        Examples
        --------
        results = await db.search(MyClass, "age", 11)

        results = await db.search(MyClass, t"name={name} AND age={age}")

        Overloads
        ---------
        search (cls, search_condition, value)
            A basic search for all objects where the attribute specified by search_condiition matches value

        search (cls, search_condition)
            Search using a string template for more complex queries.
            The search_condition must be a valid SQL search_condition.
            DO NOT include 'WHERE' in the string template.
            NOTE: If you use string template interpolation on postgres identifiers (e.g. column names)
                you must use the 'i' format specifier. Example:  t"{column_name:i} = {value}"

        Parameters
        ----------
        search_condition: Either a SQL criteria template or an attribute name
        value: Optional -- only used when the search_condition is an attribute name.


        """
        metadata = await self._get_class_metadata(cls)
        if isinstance(search_condition, str):
            # autopep8: off
            search_template = t"{search_condition:i} = {value}"
            # autopep8: on
        else:
            search_template = search_condition
        result = await self._query_handler.search(cls, metadata, search_template)
        return result

    async def select(self, cls: Type[T], select_sql: Template) -> List[T]:
        metadata = await self._get_class_metadata(cls)
        result = await self._query_handler.select(cls, metadata, select_sql)
        return result

    async def update(self, item: Identifiable):
        """
        Completely replaces an existing object in the database with item.

        :param item: The object to be stored in the database. This will completely replace any object in the database with the same id, class, and _version_schema
        :return: A list of ids that were updated (in the order returned by the Db)
        """
        metadata = await self._get_class_metadata(item.__class__)
        await self._command_handler.update(metadata, item)

    async def get(self, cls: Type[T], ids: Optional[list[int]] = None) -> list[T]:
        """
        Returns all of the objects in the database that have an id in the ids list.
        If ids is None, all objects of the given type (and _version_schema) are returned.
        """
        metadata = await self._get_class_metadata(cls)
        return await self._query_handler.get(cls, metadata, ids)

    async def delete(self, cls: Type[T], ids: list[int]) -> list[int]:
        metadata = await self._get_class_metadata(cls)
        results = await self._command_handler.delete(metadata, ids)
        return results

    async def migrate[T: (Identifiable), S:(Identifiable)](
            self,
            old_cls: Type[T],
            migrator: Callable[[List[T]], List[S]],
            delete_after_migrate: bool = False) -> List[int]:
        """
        Uses the migrator function parameter to convert all instances of old_cls to new_cls,
        saves these to the appropriate table in the database and deletes the old instances (if delete_after_migrate is true)
        """
        # Since this function just leverages other db functions, we don't need to ensure
        # the connection of class registrations -- these are done in the other db functions.
        old_items = await self.get(old_cls, [])
        new_items = migrator(old_items)
        insert_tasks = [self.insert(item) for item in new_items]
        await asyncio.gather(*insert_tasks)
        if delete_after_migrate:
            deleted_ids = await self.delete(old_cls, [item.id for item in old_items])
            # TODO: Delete the old table and class registration
            return deleted_ids
        return []

    async def get_metadata(self, cls: Type[T]):
        metadata = await self._get_class_metadata(cls)
        return metadata
